﻿using System;
using System.Collections.Generic;

namespace AssocRuleLearn
{
  class AssocRuleProgram
  {
    static void Main(string[] args)
    {
      try
      {
        Console.WriteLine("\nBegin high-confidence association rule demo\n");

        // 0-based encoded, sorted, possible dups
        List<int[]> transactions = new List<int[]>();
        transactions.Add(new int[] { 0, 3, 4, 11 });
        transactions.Add(new int[] { 1, 4, 5 });
        transactions.Add(new int[] { 3, 4, 6, 7 });
        transactions.Add(new int[] { 3, 4, 6, 7 });
        transactions.Add(new int[] { 0, 5 });
        transactions.Add(new int[] { 3, 5, 9 });
        transactions.Add(new int[] { 2, 3, 4, 7 });
        transactions.Add(new int[] { 2, 5, 8 });
        transactions.Add(new int[] { 0, 1, 2, 5, 10 });
        transactions.Add(new int[] { 2, 3, 5, 6, 7, 9 });

        Console.WriteLine("Encoded transactions are:\n");
        ShowList(transactions);

        // frequent item-sets (support = 0.30 => appear in 3 or more transactions)
        // no dups allowed
        // frequent item-sets generate candidate rules which can be evaluated against the transactions
        // these are hard-coded but in general they will be found programmatically

        List<int[]> freqItemSets = new List<int[]>();
        freqItemSets.Add(new int[] { 2, 5 });
        freqItemSets.Add(new int[] { 3, 4 });
        freqItemSets.Add(new int[] { 3, 6 });
        freqItemSets.Add(new int[] { 3, 7 });
        freqItemSets.Add(new int[] { 4, 7 });
        freqItemSets.Add(new int[] { 6, 7 });
        freqItemSets.Add(new int[] { 3, 4, 7 });
        freqItemSets.Add(new int[] { 3, 6, 7 });

        Console.WriteLine("\nFrequent item-sets (support = 0.30) are:\n");
        ShowList(freqItemSets);

        double minConPct = 0.70;
        Console.WriteLine("\nSetting minimum confidence =  " + minConPct.ToString("F3"));

        Console.WriteLine("\nStarting high-confidence rule extraction");
        List<Rule> goodRules = GetHighConfRules(freqItemSets, transactions, minConPct);

        Console.WriteLine("\nDone. Extracted association rules are:\n");
        for (int i = 0; i < goodRules.Count; ++i)
          Console.WriteLine(goodRules[i].ToString());
        
        Console.WriteLine("\nEnd high-confidence association rule demo\n");
        Console.ReadLine();
      }
      catch (Exception ex)
      {
        Console.WriteLine(ex.Message);
        Console.ReadLine();
      }
    } // Main


    static void ShowList(List<int[]> trans)
    {
      for (int i = 0; i < trans.Count; ++i)
      {
        Console.Write(i.ToString().PadLeft(2) + ": ( ");
        for (int j = 0; j < trans[i].Length; ++j)
          Console.Write(trans[i][j] + " ");
        Console.WriteLine(")");
      }
    }

    // ===============================================================================================================

    static List<Rule> GetHighConfRules(List<int[]> freqItemSets, List<int[]> trans, double minConfidencePct)
    {
      // generate candidate rules from freqItemSets, save rules that meet min confidence against transactions
      List<Rule> result = new List<Rule>();

      Dictionary<int[], int> itemSetCountDict = new Dictionary<int[], int>(); // count of item sets

      for (int i = 0; i < freqItemSets.Count; ++i) // each freq item-set generates multiple candidate rules
      {
        int[] currItemSet = freqItemSets[i]; // for clarity only
        int ctItemSet = CountInTrans(currItemSet, trans, itemSetCountDict); // needed for each candidate rule
 
        for (int len = 1; len <= currItemSet.Length - 1; ++len) // antecedent len = 1, 2, 3, . .
        {
          int[] c = NewCombination(len); // a mathematical combination

          while (c != null) // each combination makes a candidate rule
          {
            int[] ante = MakeAntecedent(currItemSet, c);
            int[] cons = MakeConsequent(currItemSet, c); // could defer this until known if needed
          
            int ctAntecendent = CountInTrans(ante, trans, itemSetCountDict); // use lookup if possible 
            double confidence = (ctItemSet * 1.0) / ctAntecendent;

            if (confidence >= minConfidencePct) // we have a winner!
            {
              Rule r = new Rule(ante, cons, confidence); 
              result.Add(r); // if freq item-sets are distinct, no dup rules ever created
            }
            c = NextCombination(c, currItemSet.Length);
          } // while each combination
        } // len each possible antecedent for curr item-set
      } // i each freq item-set

      return result;
    } // GetHighConfRules

    static int[] NewCombination(int k)
    {
      // if k = 3, return is (0 1 2). n is external somewhere
      int[] result = new int[k];
      for (int i = 0; i < result.Length; ++i)
        result[i] = i;
      return result;
    }

    static int[] NextCombination(int[] comb, int n)
    {
      // if n = 5, combination = (0 3 4) next is (1 2 3)
      // if n = 5, combination = (3 4 5) next is null
      int[] result = new int[comb.Length];
      int k = comb.Length;

      if (comb[0] == n - k) return null;
      Array.Copy(comb, result, comb.Length);
      int i = k - 1;
      while (i > 0 && result[i] == n - k + i)
        --i;
      ++result[i];
      for (int j = i; j < k - 1; ++j)
        result[j + 1] = result[j] + 1;
      return result;
    }

    static int[] MakeAntecedent(int[] itemSet, int[] comb)
    {
      // if item-set = (1 3 4 6 8) and combination = (0 2) 
      // then antecedent = (1 4)
      int[] result = new int[comb.Length];
      for (int i = 0; i < comb.Length; ++i)
      {
        int idx = comb[i];
        result[i] = itemSet[idx];
      }
      return result;
    }

    static int[] MakeConsequent(int[] itemSet, int[] comb)
    {
      // if item-set = (1 3 4 6 8) and combination = (0 2) 
      // then consequent = (3 6 8)
      int[] result = new int[itemSet.Length - comb.Length];
      int j = 0; // ptr into combination
      int p = 0; // ptr into result
      for (int i = 0; i < itemSet.Length; ++i)
      {
        if (j < comb.Length && i == comb[j]) // we are at an antecedent
          ++j; // so continue
        else
          result[p++] = itemSet[i]; // at a consequent so add it
      }
      return result;
    }

    static int CountInTrans(int[] itemSet, List<int[]> trans, Dictionary<int[], int> countDict)
    {
      // number of times itemSet occurs in transactions, using a lookup dict
      if (countDict.ContainsKey(itemSet) == true)
        return countDict[itemSet]; // use already computed count

      int ct = 0;
      for (int i = 0; i < trans.Count; ++i)
        if (IsSubsetOf(itemSet, trans[i]) == true)
          ++ct;
      countDict.Add(itemSet, ct);
      return ct;
    }

    static bool IsSubsetOf(int[] itemSet, int[] trans)
    {
      // 'trans' is an ordered transaction like [0 1 4 5 8]
      int foundIdx = -1;
      for (int j = 0; j < itemSet.Length; ++j)
      {
        foundIdx = IndexOf(trans, itemSet[j], foundIdx + 1);
        if (foundIdx == -1) return false;
      }
      return true;
    }

    static int IndexOf(int[] array, int item, int startIdx)
    {
      for (int i = startIdx; i < array.Length; ++i)
      {
        if (i > item) return -1; // i is past where the target could possibly be
        if (array[i] == item) return i;
      }
      return -1;
    }

  } // program class

  public class Rule
  {
    public int[] antecedent;
    public int[] consequent;
    public double confidence;

    public Rule(int[] antecedent, int[] consequent, double confidence)
    {
      this.antecedent = new int[antecedent.Length];
      Array.Copy(antecedent, this.antecedent, antecedent.Length);
      this.consequent = new int[consequent.Length];
      Array.Copy(consequent, this.consequent, consequent.Length);
      this.confidence = confidence;
    }

    public override string ToString() // hacked for demo data
    {
      string s = "IF ( ";
      for (int i = 0; i < antecedent.Length; ++i)
        s += antecedent[i] + " ";
      s += ")";
      s = s.PadRight(13);

      string t = " THEN ( ";
      for (int i = 0; i < consequent.Length; ++i)
        t += consequent[i] + " ";
      t += ") ";
      t = t.PadRight(17);

      return s + t + "conf = " + confidence.ToString("F2");
    }

    //public string ToShortString()
    //{
    //  string s = "";
    //  for (int i = 0; i < antecedent.Length; ++i)
    //    s += antecedent[i] + " ";
    //  s += "-> ";
    //  for (int i = 0; i < consequent.Length; ++i)
    //    s += consequent[i] + " ";
    //  return s;
    //}

  } // Rule

} // ns